/*
 * Copyright 2012-2013 M. Andersen and L. Vandenberghe.
 * Copyright 2010-2011 L. Vandenberghe.
 * Copyright 2004-2009 J. Dahl and L. Vandenberghe.
 *
 * This file is part of CVXOPT version 1.1.6.
 *
 * CVXOPT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * CVXOPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "cvxopt.h"
#include "misc.h"
#include <glpk.h>
#include <float.h>
#include <limits.h>

PyDoc_STRVAR(glpk__doc__,
    "Interface to the simplex and mixed integer LP algorithms in GLPK.\n\n"
    "The GLPK control parameters have the default values listed in \n"
    "the GLPK documentation, except for 'LPX_K_PRESOL', which is set\n"
    "to 1 and cannot be modified.  The other parameters can be\n"
    "modified by making an entry in the dictionary glpk.options.\n"
    "For example, the command glpk.options['LPX_K_MSGLEV'] = 0 turns\n"
    "off the printed output during execution of glpk.simplex().\n"
    "See the documentation at www.gnu.org/software/glpk/glpk.html for\n"
    "the list of GLPK control parameters and their default values.");

static PyObject *glpk_module;

#if PY_MAJOR_VERSION >= 3
#define PYINT_CHECK(value) PyLong_Check(value)
#define PYINT_AS_LONG(value) PyLong_AS_LONG(value)
#define PYSTRING_FROMSTRING(str) PyUnicode_FromString(str)
#else
#define PYINT_CHECK(value) PyInt_Check(value)
#define PYINT_AS_LONG(value) PyInt_AS_LONG(value)
#define PYSTRING_FROMSTRING(str) PyString_FromString(str)
#endif

static char doc_simplex[] =
    "Solves a linear program using GLPK.\n\n"
    "(status, x, z, y) = lp(c, G, h, A, b)\n"
    "(status, x, z) = lp(c, G, h)\n\n"
    "PURPOSE\n"
    "(status, x, z, y) = lp(c, G, h, A, b) solves the pair\n"
    "of primal and dual LPs\n\n"
    "    minimize    c'*x            maximize    -h'*z + b'*y\n"
    "    subject to  G*x <= h        subject to  G'*z + A'*y + c = 0\n"
    "                A*x = b                     z >= 0.\n\n"
    "(status, x, z) = lp(c, G, h) solves the pair of primal\n"
    "and dual LPs\n\n"
    "    minimize    c'*x            maximize    -h'*z \n"
    "    subject to  G*x <= h        subject to  G'*z + c = 0\n"
    "                                            z >= 0.\n\n"
    "ARGUMENTS\n"
    "c            nx1 dense 'd' matrix with n>=1\n\n"
    "G            mxn dense or sparse 'd' matrix with m>=1\n\n"
    "h            mx1 dense 'd' matrix\n\n"
    "A            pxn dense or sparse 'd' matrix with p>=0\n\n"
    "b            px1 dnese 'd' matrix\n\n"
    "status       'optimal', 'primal infeasible', 'dual infeasible' \n"
    "             or 'unknown'\n\n"
    "x            if status is 'optimal', a primal optimal solution;\n"
    "             None otherwise\n\n"
    "z,y          if status is 'optimal', the dual optimal solution;\n"
    "             None otherwise";


static PyObject *simplex(PyObject *self, PyObject *args,
    PyObject *kwrds)
{
    matrix *c, *h, *b=NULL, *x=NULL, *z=NULL, *y=NULL;
    PyObject *G, *A=NULL, *t=NULL, *param, *key, *value;
    glp_prob *lp;
    glp_smcp smcp;
    int m, n, p, i, j, k, nnz, nnzmax, *rn=NULL, *cn=NULL;
    int_t pos=0;
    double *a=NULL, val;
    char err_str[100];
#if PY_MAJOR_VERSION >= 3
    const char *keystr;
#else
    char *keystr;
#endif
    char *kwlist[] = {"c", "G", "h", "A", "b", NULL};

    if (!PyArg_ParseTupleAndKeywords(args, kwrds, "OOO|OO", kwlist, &c,
        &G, &h, &A, &b)) return NULL;

    if ((Matrix_Check(G) && MAT_ID(G) != DOUBLE) ||
        (SpMatrix_Check(G) && SP_ID(G) != DOUBLE) ||
        (!Matrix_Check(G) && !SpMatrix_Check(G))){
        PyErr_SetString(PyExc_TypeError, "G must be a 'd' matrix");
        return NULL;
    }
    if ((m = Matrix_Check(G) ? MAT_NROWS(G) : SP_NROWS(G)) <= 0)
        err_p_int("m");
    if ((n = Matrix_Check(G) ? MAT_NCOLS(G) : SP_NCOLS(G)) <= 0)
        err_p_int("n");

    if (!Matrix_Check(h) || h->id != DOUBLE) err_dbl_mtrx("h");
    if (h->nrows != m || h->ncols != 1){
        PyErr_SetString(PyExc_ValueError, "incompatible dimensions");
        return NULL;
    }

    if (A){
        if ((Matrix_Check(A) && MAT_ID(A) != DOUBLE) ||
            (SpMatrix_Check(A) && SP_ID(A) != DOUBLE) ||
            (!Matrix_Check(A) && !SpMatrix_Check(A))){
                PyErr_SetString(PyExc_ValueError, "A must be a dense "
                    "'d' matrix or a general sparse matrix");
                return NULL;
	}
        if ((p = Matrix_Check(A) ? MAT_NROWS(A) : SP_NROWS(A)) < 0)
            err_p_int("p");
        if ((Matrix_Check(A) ? MAT_NCOLS(A) : SP_NCOLS(A)) != n){
            PyErr_SetString(PyExc_ValueError, "incompatible "
                "dimensions");
            return NULL;
	}
    }
    else p = 0;

    if (b && (!Matrix_Check(b) || b->id != DOUBLE)) err_dbl_mtrx("b");
    if ((b && (b->nrows != p || b->ncols != 1)) || (!b && p !=0 )){
        PyErr_SetString(PyExc_ValueError, "incompatible dimensions");
        return NULL;
    }

    lp = glp_create_prob();
    glp_add_rows(lp, m+p);
    glp_add_cols(lp, n);

    for (i=0; i<n; i++){
        glp_set_obj_coef(lp, i+1, MAT_BUFD(c)[i]);
        glp_set_col_bnds(lp, i+1, GLP_FR, 0.0, 0.0);
    }
    for (i=0; i<m; i++)
        glp_set_row_bnds(lp, i+1, GLP_UP, 0.0, MAT_BUFD(h)[i]);
    for (i=0; i<p; i++)
        glp_set_row_bnds(lp, i+m+1, GLP_FX, MAT_BUFD(b)[i],
            MAT_BUFD(b)[i]);

    nnzmax = (SpMatrix_Check(G) ? SP_NNZ(G) : m*n ) +
        ((A && SpMatrix_Check(A)) ? SP_NNZ(A) : p*n);
    a = (double *) calloc(nnzmax+1, sizeof(double));
    rn = (int *) calloc(nnzmax+1, sizeof(int));
    cn = (int *) calloc(nnzmax+1, sizeof(int));
    if (!a || !rn || !cn){
        free(a);  free(rn);  free(cn);  glp_delete_prob(lp);
        return PyErr_NoMemory();
    }

    nnz = 0;
    if (SpMatrix_Check(G)) {
        for (j=0; j<n; j++) for (k=SP_COL(G)[j]; k<SP_COL(G)[j+1]; k++)
            if ((val = SP_VALD(G)[k]) != 0.0){
                a[1+nnz] = val;
                rn[1+nnz] = SP_ROW(G)[k]+1;
                cn[1+nnz] = j+1;
                nnz++;
            }
    }
    else for (j=0; j<n; j++) for (i=0; i<m; i++)
        if ((val = MAT_BUFD(G)[i+j*m]) != 0.0){
            a[1+nnz] = val;
            rn[1+nnz] = i+1;
            cn[1+nnz] = j+1;
            nnz++;
        }

    if (A && SpMatrix_Check(A)){
        for (j=0; j<n; j++) for (k=SP_COL(A)[j]; k<SP_COL(A)[j+1]; k++)
            if ((val = SP_VALD(A)[k]) != 0.0){
                a[1+nnz] = val;
                rn[1+nnz] = m+SP_ROW(A)[k]+1;
                cn[1+nnz] = j+1;
                nnz++;
            }
    }
    else for (j=0; j<n; j++) for (i=0; i<p; i++)
        if ((val = MAT_BUFD(A)[i+j*p]) != 0.0){
            a[1+nnz] = val;
            rn[1+nnz] = m+i+1;
            cn[1+nnz] = j+1;
            nnz++;
        }

    glp_load_matrix(lp, nnz, rn, cn, a);
    free(rn);  free(cn);  free(a);

    if (!(t = PyTuple_New(A ? 4 : 3))){
        glp_delete_prob(lp);
        return PyErr_NoMemory();
    }

    if (!(param = PyObject_GetAttrString(glpk_module, "options"))
        || !PyDict_Check(param)){
            glp_delete_prob(lp);
            PyErr_SetString(PyExc_AttributeError,
                "missing glpk.options dictionary");
            return NULL;
    }

    glp_init_smcp(&smcp);
    while (PyDict_Next(param, &pos, &key, &value)){
#if PY_MAJOR_VERSION >= 3
        if (PyUnicode_Check(key)){
            keystr = _PyUnicode_AsString(key);
#else
        if ((keystr = PyString_AsString(key))){
#endif
            if (!strcmp("LPX_K_MSGLEV", keystr)){
                switch (PYINT_AS_LONG(value)){
                    case 0: smcp.msg_lev = GLP_MSG_OFF; break;
                    case 1: smcp.msg_lev = GLP_MSG_ERR; break;
                    case 2: smcp.msg_lev = GLP_MSG_ON; break;
                    case 3: smcp.msg_lev = GLP_MSG_ALL; break;
                    default: goto error;
                }
            } else if (!strcmp("LPX_K_SCALE", keystr)){
                switch(PYINT_AS_LONG(value)){
                    case 0: glp_unscale_prob(lp); break;
                    case 1: glp_scale_prob(lp, GLP_SF_EQ); break;
                    case 2: glp_scale_prob(lp, GLP_SF_GM); break;
                    case 3: glp_scale_prob(lp, GLP_SF_GM | GLP_SF_EQ); break;
                    default: goto error;
                }
            } else if (!strcmp("LPX_K_DUAL", keystr)){
                switch(PYINT_AS_LONG(value)){
                    case 0: smcp.meth = GLP_PRIMAL; break;
                    case 1: smcp.meth = GLP_DUAL; break;
                    default: goto error;
                }
            } else if (!strcmp("LPX_K_PRICE", keystr)){
                switch(PYINT_AS_LONG(value)){
                    case 0: smcp.pricing = GLP_PT_STD; break;
                    case 1: smcp.pricing = GLP_PT_PSE; break;
                    default: goto error;
                }
            } else if (!strcmp("LPX_K_RELAX", keystr)){
                double relax = PyFloat_AsDouble(value);
                if (relax < 0.0 || relax > 1.0)
                    goto error;
                smcp.r_test = (relax == 0.0) ? GLP_RT_STD : GLP_RT_HAR;
            } else if (!strcmp("LPX_K_TOLBND", keystr)){
                smcp.tol_bnd = PyFloat_AsDouble(value);
                if (smcp.tol_bnd < DBL_EPSILON || smcp.tol_bnd > 0.001)
                    goto error;
            } else if (!strcmp("LPX_K_TOLDJ", keystr)){
                smcp.tol_dj = PyFloat_AsDouble(value);
                if (smcp.tol_dj < DBL_EPSILON || smcp.tol_dj > 0.001)
                    goto error;
            } else if (!strcmp("LPX_K_TOLPIV", keystr)){
                smcp.tol_piv = PyFloat_AsDouble(value);
                if (smcp.tol_piv < DBL_EPSILON || smcp.tol_piv > 0.001)
                    goto error;
            } else if (!strcmp("LPX_K_ROUND", keystr)){
                /* Ignored */
            } else if (!strcmp("LPX_K_OBJLL", keystr)){
                smcp.obj_ll = PyFloat_AsDouble(value);
            } else if (!strcmp("LPX_K_OBJUL", keystr)){
                smcp.obj_ul = PyFloat_AsDouble(value);
            } else if (!strcmp("LPX_K_ITLIM", keystr)){
                smcp.it_lim = PYINT_AS_LONG(value);
                if (smcp.it_lim < 0)
                    smcp.it_lim = INT_MAX;
            } else if (!strcmp("LPX_K_ITCNT", keystr)){
                /* Ignored */
            } else if (!strcmp("LPX_K_TMLIM", keystr)){
                smcp.tm_lim = (int)(PyFloat_AsDouble(value) * 1000.0);
                if (smcp.tm_lim < 0)
                    smcp.tm_lim = INT_MAX;
            } else if (!strcmp("LPX_K_OUTFRQ", keystr)){
                smcp.out_frq = PYINT_AS_LONG(value);
                if (smcp.out_frq <= 0)
                    goto error;
            } else if (!strcmp("LPX_K_OUTDLY", keystr)){
                smcp.out_dly = (int)(PyFloat_AsDouble(value) * 1000.0);
            } else if (!strcmp("LPX_K_BRANCH", keystr)){
                /* Ignored */
            } else if (!strcmp("LPX_K_BTRACK", keystr)){
                /* Ignored */
            } else if (!strcmp("LPX_K_TOLINT", keystr)){
                /* Ignored */
            } else if (!strcmp("LPX_K_TOLOBJ", keystr)){
                /* Ignored */
            } else if (!strcmp("LPX_K_MPSINFO", keystr)){
                /* Ignored */
            } else if (!strcmp("LPX_K_MPSOBJ", keystr)){
                /* Ignored */
            } else if (!strcmp("LPX_K_MPSORIG", keystr)){
                /* Ignored */
            } else if (!strcmp("LPX_K_MPSWIDE", keystr)){
                /* Ignored */
            } else if (!strcmp("LPX_K_MPSFREE", keystr)){
                /* Ignored */
            } else if (!strcmp("LPX_K_MPSSKIP", keystr)){
                /* Ignored */
            } else if (!strcmp("LPX_K_LPTORIG", keystr)){
                /* Ignored */
            } else if (!strcmp("LPX_K_PRESOL", keystr)){
                if (PYINT_AS_LONG(value) != 1)
                    PyErr_Warn(PyExc_UserWarning, "ignoring value of "
                        "GLPK parameter 'LPX_K_PRESOL'");
            } else if (!strcmp("LPX_K_BINARIZE", keystr)){
                /* Ignored */
            } else if (!strcmp("LPX_K_USECUTS", keystr)){
                /* Ignored */
            } else if (!strcmp("LPX_K_BFTYPE", keystr)){
                glp_bfcp bfcp;

                glp_get_bfcp(lp, &bfcp);
                switch(PYINT_AS_LONG(value)){
                    case 1: bfcp.type = GLP_BF_FT; break;
                    case 2: bfcp.type = GLP_BF_BG; break;
                    case 3: bfcp.type = GLP_BF_GR; break;
                    default: goto error;
                }
                glp_set_bfcp(lp, &bfcp);
            } else if (!strcmp("LPX_K_MIPGAP", keystr)){
                /* Ignored */
            } else {
                sprintf(err_str, "unknown GLPK parameter: %-.20s", keystr);
                PyErr_SetString(PyExc_ValueError, err_str);
                glp_delete_prob(lp);
                Py_DECREF(param);
                return NULL;
	    }
        }
    }
    smcp.presolve = GLP_ON;
    Py_DECREF(param);

    switch (glp_simplex(lp, &smcp)){

        case 0:

            x = (matrix *) Matrix_New(n,1,DOUBLE);
            z = (matrix *) Matrix_New(m,1,DOUBLE);
            if (A) y = (matrix *) Matrix_New(p,1,DOUBLE);
            if (!x || !z || (A && !y)){
                Py_XDECREF(x);
                Py_XDECREF(z);
                Py_XDECREF(y);
                Py_XDECREF(t);
                glp_delete_prob(lp);
                return PyErr_NoMemory();
            }

            PyTuple_SET_ITEM(t, 0, (PyObject *)PYSTRING_FROMSTRING("optimal"));

            for (i=0; i<n; i++)
                MAT_BUFD(x)[i] = glp_get_col_prim(lp, i+1);
            PyTuple_SET_ITEM(t, 1, (PyObject *) x);

            for (i=0; i<m; i++)
                MAT_BUFD(z)[i] = -glp_get_row_dual(lp, i+1);
            PyTuple_SET_ITEM(t, 2, (PyObject *) z);

            if (A){
                for (i=0; i<p; i++)
                    MAT_BUFD(y)[i] = -glp_get_row_dual(lp, m+i+1);
                PyTuple_SET_ITEM(t, 3, (PyObject *) y);
            }

            glp_delete_prob(lp);
            return (PyObject *) t;

        case GLP_ENOPFS:

            PyTuple_SET_ITEM(t, 0, (PyObject *)
                PYSTRING_FROMSTRING("primal infeasible"));
            break;

        case GLP_ENODFS:

            PyTuple_SET_ITEM(t, 0, (PyObject *)
                PYSTRING_FROMSTRING("dual infeasible"));
            break;

        default:

            PyTuple_SET_ITEM(t, 0, (PyObject *)
                PYSTRING_FROMSTRING("unknown"));
    }

    glp_delete_prob(lp);

    PyTuple_SET_ITEM(t, 1, Py_BuildValue(""));
    PyTuple_SET_ITEM(t, 2, Py_BuildValue(""));
    if (A) PyTuple_SET_ITEM(t, 3, Py_BuildValue(""));

    return (PyObject *) t;

error:
    sprintf(err_str, "invalid value for GLPK parameter: %-.20s", keystr);
    PyErr_SetString(PyExc_ValueError, err_str);
    glp_delete_prob(lp);
    Py_DECREF(param);
    return NULL;
}



static char doc_integer[] =
    "Solves a mixed integer linear program using GLPK.\n\n"
    "(status, x) = ilp(c, G, h, A, b, I, B)\n\n"
    "PURPOSE\n"
    "Solves the mixed integer linear programming problem\n\n"
    "    minimize    c'*x\n"
    "    subject to  G*x <= h\n"
    "                A*x = b\n"
    "                x[I] are all integer\n"
    "                x[B] are all binary\n\n"
    "ARGUMENTS\n"
    "c            nx1 dense 'd' matrix with n>=1\n\n"
    "G            mxn dense or sparse 'd' matrix with m>=1\n\n"
    "h            mx1 dense 'd' matrix\n\n"
    "A            pxn dense or sparse 'd' matrix with p>=0\n\n"
    "b            px1 dense 'd' matrix\n\n"
    "I            set with indices of integer variables\n\n"
    "B            set with indices of binary variables\n\n"
    "status       'optimal', 'primal infeasible', 'dual infeasible', \n"
    "             'invalid MIP formulation', 'maxiters exceeded', \n"
    "             'time limit exceeded', 'unknown'\n\n"
    "x            a (sub-)optimal solution if status is 'optimal' or \n"
    "             'time limit exceeded'; None otherwise";

static PyObject *integer(PyObject *self, PyObject *args,
    PyObject *kwrds)
{
    matrix *c, *h, *b=NULL, *x=NULL;
    PyObject *G, *A=NULL, *IntSet=NULL, *BinSet = NULL;
    PyObject *t=NULL, *param, *key, *value;
    glp_prob *lp;
    glp_iocp iocp;
    int m, n, p, i, j, k, nnz, nnzmax, *rn=NULL, *cn=NULL;
    int_t pos=0;
    double *a=NULL, val;
    char err_str[100]; 
#if PY_MAJOR_VERSION >= 3
    const char *keystr;
#else
    char *keystr;
#endif
    char *kwlist[] = {"c", "G", "h", "A", "b", "I", "B", NULL};

    if (!PyArg_ParseTupleAndKeywords(args, kwrds, "OOO|OOOO", kwlist, &c,
	    &G, &h, &A, &b, &IntSet, &BinSet)) return NULL;

    if ((Matrix_Check(G) && MAT_ID(G) != DOUBLE) ||
        (SpMatrix_Check(G) && SP_ID(G) != DOUBLE) ||
        (!Matrix_Check(G) && !SpMatrix_Check(G))){
        PyErr_SetString(PyExc_TypeError, "G must be a 'd' matrix");
        return NULL;
    }
    if ((m = Matrix_Check(G) ? MAT_NROWS(G) : SP_NROWS(G)) <= 0)
        err_p_int("m");
    if ((n = Matrix_Check(G) ? MAT_NCOLS(G) : SP_NCOLS(G)) <= 0)
        err_p_int("n");

    if (!Matrix_Check(h) || h->id != DOUBLE) err_dbl_mtrx("h");
    if (h->nrows != m || h->ncols != 1){
        PyErr_SetString(PyExc_ValueError, "incompatible dimensions");
        return NULL;
    }

    if (A){
        if ((Matrix_Check(A) && MAT_ID(A) != DOUBLE) ||
            (SpMatrix_Check(A) && SP_ID(A) != DOUBLE) ||
            (!Matrix_Check(A) && !SpMatrix_Check(A))){
                PyErr_SetString(PyExc_ValueError, "A must be a dense "
                    "'d' matrix or a general sparse matrix");
                return NULL;
	}
        if ((p = Matrix_Check(A) ? MAT_NROWS(A) : SP_NROWS(A)) < 0)
            err_p_int("p");
        if ((Matrix_Check(A) ? MAT_NCOLS(A) : SP_NCOLS(A)) != n){
            PyErr_SetString(PyExc_ValueError, "incompatible "
                "dimensions");
            return NULL;
	}
    }
    else p = 0;

    if (b && (!Matrix_Check(b) || b->id != DOUBLE)) err_dbl_mtrx("b");
    if ((b && (b->nrows != p || b->ncols != 1)) || (!b && p !=0 )){
        PyErr_SetString(PyExc_ValueError, "incompatible dimensions");
        return NULL;
    }

    if ((IntSet) && (!PyAnySet_Check(IntSet)))
      PY_ERR_TYPE("invalid integer index set");

    if ((BinSet) && (!PyAnySet_Check(BinSet)))
      PY_ERR_TYPE("invalid binary index set");

    lp = glp_create_prob();
    glp_add_rows(lp, m+p);
    glp_add_cols(lp, n);

    for (i=0; i<n; i++){
        glp_set_obj_coef(lp, i+1, MAT_BUFD(c)[i]);
        glp_set_col_bnds(lp, i+1, GLP_FR, 0.0, 0.0);
    }
    for (i=0; i<m; i++)
        glp_set_row_bnds(lp, i+1, GLP_UP, 0.0, MAT_BUFD(h)[i]);
    for (i=0; i<p; i++)
        glp_set_row_bnds(lp, i+m+1, GLP_FX, MAT_BUFD(b)[i],
            MAT_BUFD(b)[i]);

    nnzmax = (SpMatrix_Check(G) ? SP_NNZ(G) : m*n ) +
        ((A && SpMatrix_Check(A)) ? SP_NNZ(A) : p*n);
    a = (double *) calloc(nnzmax+1, sizeof(double));
    rn = (int *) calloc(nnzmax+1, sizeof(int));
    cn = (int *) calloc(nnzmax+1, sizeof(int));
    if (!a || !rn || !cn){
        free(a);  free(rn);  free(cn);  glp_delete_prob(lp);
        return PyErr_NoMemory();
    }

    nnz = 0;
    if (SpMatrix_Check(G)) {
        for (j=0; j<n; j++) for (k=SP_COL(G)[j]; k<SP_COL(G)[j+1]; k++)
            if ((val = SP_VALD(G)[k]) != 0.0){
                a[1+nnz] = val;
                rn[1+nnz] = SP_ROW(G)[k]+1;
                cn[1+nnz] = j+1;
                nnz++;
            }
    }
    else for (j=0; j<n; j++) for (i=0; i<m; i++)
        if ((val = MAT_BUFD(G)[i+j*m]) != 0.0){
            a[1+nnz] = val;
            rn[1+nnz] = i+1;
            cn[1+nnz] = j+1;
            nnz++;
        }

    if (A && SpMatrix_Check(A)){
        for (j=0; j<n; j++) for (k=SP_COL(A)[j]; k<SP_COL(A)[j+1]; k++)
            if ((val = SP_VALD(A)[k]) != 0.0){
                a[1+nnz] = val;
                rn[1+nnz] = m+SP_ROW(A)[k]+1;
                cn[1+nnz] = j+1;
                nnz++;
            }
    }
    else for (j=0; j<n; j++) for (i=0; i<p; i++)
        if ((val = MAT_BUFD(A)[i+j*p]) != 0.0){
            a[1+nnz] = val;
            rn[1+nnz] = m+i+1;
            cn[1+nnz] = j+1;
            nnz++;
        }

    glp_load_matrix(lp, nnz, rn, cn, a);
    free(rn);  free(cn);  free(a);

    if (!(t = PyTuple_New(2))) {
        glp_delete_prob(lp);
        return PyErr_NoMemory();
    }

    if (!(param = PyObject_GetAttrString(glpk_module, "options"))
        || !PyDict_Check(param)){
            glp_delete_prob(lp);
            PyErr_SetString(PyExc_AttributeError,
                "missing glpk.options dictionary");
            return NULL;
    }

    glp_init_iocp(&iocp);
    while (PyDict_Next(param, &pos, &key, &value)) {
#if PY_MAJOR_VERSION >= 3
        if ((PyUnicode_Check(key)) && (keystr = PyUnicode_AS_DATA(key))){
#else
        if ((keystr = PyString_AsString(key))){
#endif
            if (!strcmp("LPX_K_MSGLEV", keystr)){
                switch (PYINT_AS_LONG(value)){
                    case 0: iocp.msg_lev = GLP_MSG_OFF; break;
                    case 1: iocp.msg_lev = GLP_MSG_ERR; break;
                    case 2: iocp.msg_lev = GLP_MSG_ON; break;
                    case 3: iocp.msg_lev = GLP_MSG_ALL; break;
                    default: goto error;
                }
            } else if (!strcmp("LPX_K_SCALE", keystr)){
                switch(PYINT_AS_LONG(value)){
                    case 0: glp_unscale_prob(lp); break;
                    case 1: glp_scale_prob(lp, GLP_SF_EQ); break;
                    case 2: glp_scale_prob(lp, GLP_SF_GM); break;
                    case 3: glp_scale_prob(lp, GLP_SF_GM | GLP_SF_EQ); break;
                    default: goto error;
                }
            } else if (!strcmp("LPX_K_DUAL", keystr)){
                /* Ignored */
            } else if (!strcmp("LPX_K_PRICE", keystr)){
                /* Ignored */
            } else if (!strcmp("LPX_K_RELAX", keystr)){
                /* Ignored */
            } else if (!strcmp("LPX_K_TOLBND", keystr)){
                /* Ignored */
            } else if (!strcmp("LPX_K_TOLDJ", keystr)){
                /* Ignored */
            } else if (!strcmp("LPX_K_TOLPIV", keystr)){
                /* Ignored */
            } else if (!strcmp("LPX_K_ROUND", keystr)){
                /* Ignored */
            } else if (!strcmp("LPX_K_OBJLL", keystr)){
                /* Ignored */
            } else if (!strcmp("LPX_K_OBJUL", keystr)){
                /* Ignored */
            } else if (!strcmp("LPX_K_ITLIM", keystr)){
                /* Ignored */
            } else if (!strcmp("LPX_K_ITCNT", keystr)){
                /* Ignored */
            } else if (!strcmp("LPX_K_TMLIM", keystr)){
                iocp.tm_lim = (int)(PyFloat_AsDouble(value) * 1000.0);
                if (iocp.tm_lim < 0)
                    iocp.tm_lim = INT_MAX;
            } else if (!strcmp("LPX_K_OUTFRQ", keystr)){
                /* Ignored */
            } else if (!strcmp("LPX_K_OUTDLY", keystr)){
                /* Ignored */
            } else if (!strcmp("LPX_K_BRANCH", keystr)){
                switch(PYINT_AS_LONG(value)){
                    case 0: iocp.br_tech = GLP_BR_FFV; break;
                    case 1: iocp.br_tech = GLP_BR_LFV; break;
                    case 2: iocp.br_tech = GLP_BR_DTH; break;
                    case 3: iocp.br_tech = GLP_BR_MFV; break;
                    default: goto error;
                }
            } else if (!strcmp("LPX_K_BTRACK", keystr)){
                switch(PYINT_AS_LONG(value)){
                    case 0: iocp.bt_tech = GLP_BT_DFS; break;
                    case 1: iocp.bt_tech = GLP_BT_BFS; break;
                    case 2: iocp.bt_tech = GLP_BT_BPH; break;
                    case 3: iocp.bt_tech = GLP_BT_BLB; break;
                    default: goto error;
                }
            } else if (!strcmp("LPX_K_TOLINT", keystr)){
                iocp.tol_int = PyFloat_AsDouble(value);
                if (iocp.tol_int < DBL_EPSILON || iocp.tol_int > 0.001)
                    goto error;
            } else if (!strcmp("LPX_K_TOLOBJ", keystr)){
                iocp.tol_obj = PyFloat_AsDouble(value);
                if (iocp.tol_obj < DBL_EPSILON || iocp.tol_obj > 0.001)
                    goto error;
            } else if (!strcmp("LPX_K_MPSINFO", keystr)){
                /* Ignored */
            } else if (!strcmp("LPX_K_MPSOBJ", keystr)){
                /* Ignored */
            } else if (!strcmp("LPX_K_MPSORIG", keystr)){
                /* Ignored */
            } else if (!strcmp("LPX_K_MPSWIDE", keystr)){
                /* Ignored */
            } else if (!strcmp("LPX_K_MPSFREE", keystr)){
                /* Ignored */
            } else if (!strcmp("LPX_K_MPSSKIP", keystr)){
                /* Ignored */
            } else if (!strcmp("LPX_K_LPTORIG", keystr)){
                /* Ignored */
            } else if (!strcmp("LPX_K_PRESOL", keystr)){
                if (PYINT_AS_LONG(value) != 1)
                    PyErr_Warn(PyExc_UserWarning, "ignoring value of "
                        "GLPK parameter 'LPX_K_PRESOL'");
            } else if (!strcmp("LPX_K_BINARIZE", keystr)){
                iocp.binarize = (PYINT_AS_LONG(value) == 0) ? GLP_OFF : GLP_ON;
            } else if (!strcmp("LPX_K_USECUTS", keystr)){
                int cuts = PYINT_AS_LONG(value);
                if (cuts & ~0xFF)
                    goto error;
                iocp.cov_cuts = (cuts & 1) ? GLP_ON : GLP_OFF;
                iocp.clq_cuts = (cuts & 2) ? GLP_ON : GLP_OFF;
                iocp.gmi_cuts = (cuts & 4) ? GLP_ON : GLP_OFF;
                iocp.mir_cuts = (cuts & 8) ? GLP_ON : GLP_OFF;
            } else if (!strcmp("LPX_K_BFTYPE", keystr)){
                glp_bfcp bfcp;

                glp_get_bfcp(lp, &bfcp);
                switch(PYINT_AS_LONG(value)){
                    case 1: bfcp.type = GLP_BF_FT; break;
                    case 2: bfcp.type = GLP_BF_BG; break;
                    case 3: bfcp.type = GLP_BF_GR; break;
                    default: goto error;
                }
                glp_set_bfcp(lp, &bfcp);
            } else if (!strcmp("LPX_K_MIPGAP", keystr)){
                iocp.mip_gap = PyFloat_AsDouble(value);
                if (iocp.mip_gap < 0.0)
                    goto error;
            } else {
                sprintf(err_str, "unknown GLPK parameter: %-.20s", keystr);
                PyErr_SetString(PyExc_ValueError, err_str);
                glp_delete_prob(lp);
                Py_DECREF(param);
                return NULL;
	    }
        }
    }
    iocp.presolve = GLP_ON;
    Py_DECREF(param);

    if (IntSet) {
      PyObject *iter = PySequence_Fast(IntSet, "Critical error: not sequence");

      for (i=0; i<PySet_GET_SIZE(IntSet); i++) {

	PyObject *tmp = PySequence_Fast_GET_ITEM(iter, i);
	if (!PYINT_CHECK(tmp)) {
	  glp_delete_prob(lp);
	  Py_DECREF(iter);
	  PY_ERR_TYPE("non-integer element in I");
	}
	int k = PYINT_AS_LONG(tmp);
	if ((k < 0) || (k >= n)) {
	  glp_delete_prob(lp);
	  Py_DECREF(iter);
	  PY_ERR(PyExc_IndexError, "index element out of range in I");
	}
	glp_set_col_kind(lp, k+1, GLP_IV);
      }

      Py_DECREF(iter);
    }

    if (BinSet) {
      PyObject *iter = PySequence_Fast(BinSet, "Critical error: not sequence");

      for (i=0; i<PySet_GET_SIZE(BinSet); i++) {

	PyObject *tmp = PySequence_Fast_GET_ITEM(iter, i);
	if (!PYINT_CHECK(tmp)) {
	  glp_delete_prob(lp);
	  Py_DECREF(iter);
	  PY_ERR_TYPE("non-binary element in I");
	}
	int k = PYINT_AS_LONG(tmp);
	if ((k < 0) || (k >= n)) {
	  glp_delete_prob(lp);
	  Py_DECREF(iter);
	  PY_ERR(PyExc_IndexError, "index element out of range in B");
	}
	glp_set_col_kind(lp, k+1, GLP_BV);
      }

      Py_DECREF(iter);

    }


    switch (glp_intopt(lp, &iocp)){

        case 0:

            x = (matrix *) Matrix_New(n,1,DOUBLE);
            if (!x) {
                Py_XDECREF(t);
                glp_delete_prob(lp);
                return PyErr_NoMemory();
            }
            PyTuple_SET_ITEM(t, 0, (PyObject *)PYSTRING_FROMSTRING("optimal"));

            for (i=0; i<n; i++)
                MAT_BUFD(x)[i] = glp_mip_col_val(lp, i+1);
            PyTuple_SET_ITEM(t, 1, (PyObject *) x);

            glp_delete_prob(lp);
            return (PyObject *) t;

        case GLP_ETMLIM:

            x = (matrix *) Matrix_New(n,1,DOUBLE);
            if (!x) {
                Py_XDECREF(t);
                glp_delete_prob(lp);
                return PyErr_NoMemory();
            }
            PyTuple_SET_ITEM(t, 0, (PyObject *)
                PYSTRING_FROMSTRING("time limit exceeded"));

            for (i=0; i<n; i++)
                MAT_BUFD(x)[i] = glp_mip_col_val(lp, i+1);
            PyTuple_SET_ITEM(t, 1, (PyObject *) x);

            glp_delete_prob(lp);
            return (PyObject *) t;


        case GLP_EBADB:
        case GLP_ECOND:
        case GLP_EBOUND:
            PyTuple_SET_ITEM(t, 0, (PyObject *)
                PYSTRING_FROMSTRING("invalid MIP formulation"));
            break;

	case GLP_ENOPFS:
            PyTuple_SET_ITEM(t, 0, (PyObject *)
                PYSTRING_FROMSTRING("primal infeasible"));
            break;

	case GLP_ENODFS:

            PyTuple_SET_ITEM(t, 0, (PyObject *)
                PYSTRING_FROMSTRING("dual infeasible"));
            break;

        case GLP_EITLIM:

            PyTuple_SET_ITEM(t, 0, (PyObject *)
                PYSTRING_FROMSTRING("maxiters exceeded"));
            break;

	case GLP_ESING:

            PyTuple_SET_ITEM(t, 0, (PyObject *)
                PYSTRING_FROMSTRING("singular or ill-conditioned basis"));
            break;


        default:

            PyTuple_SET_ITEM(t, 0, (PyObject *)
                PYSTRING_FROMSTRING("unknown"));
    }

    glp_delete_prob(lp);

    PyTuple_SET_ITEM(t, 1, Py_BuildValue(""));
    return (PyObject *) t;

error:
    sprintf(err_str, "invalid value for GLPK parameter: %-.20s", keystr);
    PyErr_SetString(PyExc_ValueError, err_str);
    glp_delete_prob(lp);
    Py_DECREF(param);
    return NULL;
}


static PyMethodDef glpk_functions[] = {
    {"lp", (PyCFunction) simplex, METH_VARARGS|METH_KEYWORDS,
        doc_simplex},
    {"ilp", (PyCFunction) integer, METH_VARARGS|METH_KEYWORDS,
        doc_integer},
    {NULL}  /* Sentinel */
};

#if PY_MAJOR_VERSION >= 3

static PyModuleDef glpk_module_def = {
    PyModuleDef_HEAD_INIT,
    "glpk",
    glpk__doc__,
    -1,
    glpk_functions,
    NULL, NULL, NULL, NULL
};

PyMODINIT_FUNC PyInit_glpk(void)
{
  if (!(glpk_module = PyModule_Create(&glpk_module_def))) return NULL;
  PyModule_AddObject(glpk_module, "options", PyDict_New());
  if (import_cvxopt() < 0) return NULL;
  return glpk_module;
}

#else

PyMODINIT_FUNC initglpk(void)
{
    glpk_module = Py_InitModule3("cvxopt.glpk", glpk_functions, 
        glpk__doc__);
    PyModule_AddObject(glpk_module, "options", PyDict_New());
    if (import_cvxopt() < 0) return;
}

#endif
